program mSyncFos;
(*
**
** Interface to Synchronet BBS FOSSIL driver, for TelSrv
**
** Created : 21-Oct-2001
** Last update : 29-May-2004
** Version: 4.05.29
**
** (c)2001-2002 by Maarten Bekers - syncfos@elebbs.com.
** Placed in the public domain.
** Modified by Rick Parrish
**
** Interfaces to code written which is (c) Rob Swindell.
**
*)
uses
  Telnet, Windows;

(*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-*)

const
  WIN9X                     = VER_PLATFORM_WIN32_WINDOWS;
  WINNT                     = VER_PLATFORM_WIN32_NT;
  PROG_VER                  = 'mSyncFos v4.05.29';
  XTRN_IO_BUF_LEN           = 5000;
  SBBSEXEC_MODE_FOSSIL      = $0000;
  SBBSEXEC_IOCTL_START      = $8001;
  SBBSEXEC_IOCTL_COMPLETE   = $8002;
  SBBSEXEC_IOCTL_READ       = $8003;
  SBBSEXEC_IOCTL_WRITE      = $8004;
  SBBSEXEC_IOCTL_DISCONNECT = $8005;
  SBBSEXEC_IOCTL_STOP       = $8006;

type
  TBuf = Array[0..XTRN_IO_BUF_LEN - 1] of Char;
  TOpenVxDHandle = function(ASource: THandle): THandle; stdcall;
  TSBBSEXEC_START_T = Record
    Mode: Cardinal;
    Event: THandle;
  end;

var
  // Com object
  ComObj: PTelnetObj;

  // OS version
  OS: Integer;

  // Setting variables
  Command: String;
  LBY: LongInt;
  Node: LongInt;
  Sock: LongInt;

  // File names
  EnvFile: String;
  RetFile: String;

  // 9x handles
  Start: TSBBSEXEC_START_T;
  StartEvent: THandle;
  VM: Cardinal;
  VXD: THandle;

  // NT handles
  HungUpEvent: THandle;
  ReadSlot: THandle;
  WriteSlot: THandle;

(*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-*)

// Com Read Procedure
procedure ComReadProc(var TempPtr: Pointer);
begin
     ComObj^.Com_ReadProc(TempPtr);
end;

// Com Write Procedure
procedure ComWriteProc(var TempPtr: Pointer);
begin
     ComObj^.Com_WriteProc(TempPtr);
end;

(*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-*)

function ExtractFilePath(AFileName: String): String;
var
  I: Integer;
begin
  Result := AFileName;

  for I := Length(AFileName) downto 1 do
  begin
       if (Pos(AFileName[I], ':\') > 0) then
       begin
            Result := Copy(AFileName, 1, I);
            break;
       end;
  end;
end;

function ExtractShortPathName(AFileName: String): String;
var
   Buffer: Array[0..MAX_PATH - 1] of Char;
begin
     SetString(Result, Buffer, GetShortPathName(PChar(AFileName), Buffer, SizeOf(Buffer)));
end;

function IntToStr(AInt: Integer): String;
begin
     Str(AInt, Result);
end;

function StrToIntDef(AStr: String; ADefault: Integer): Integer;
var
   E: Integer;
begin
     Val(AStr, Result, E);
     if E <> 0 then
        Result := ADefault;
end;

(*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-*)

// Returns true if a carrier is present or false if the user hungup
function CheckCarrier9x: Boolean;
var
  NumRead: Cardinal;
begin
     Result := ComObj^.Com_Carrier;

     // If the user hungup, perform disconnect operation
     if Not(Result) then
        DeviceIOControl(VXD,
                        SBBSEXEC_IOCTL_DISCONNECT,
                        @VM,
                        SizeOf(VM),
                        nil,
                        0,
                        NumRead,
                        nil);
end;

// Returns true if a carrier is present or false if the user hungup
function CheckCarrierNT: Boolean;
begin
     Result := ComObj^.Com_Carrier;

     // If the user hungup, set the hungup event
     if Not(Result) then
        SetEvent(HungUpEvent);
end;

// Returns true if a carrier is present or false if the user hungup
function CheckCarrier: Boolean;
begin
     if (OS = WIN9X) then
        Result := CheckCarrier9x
     else
     if (OS = WINNT) then
        Result := CheckCarrierNT
     else
        Result := False;
end;

(*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-*)

// Check to see if the program has anything to send to the user
procedure CheckRx9x;
var
  Buf: TBuf;
  NumRead: Cardinal;
  NumWritten: Integer;
begin
     // Check to see if any data is waiting
     DeviceIOControl(VXD,
                     SBBSEXEC_IOCTL_READ,
                     @VM,
                     SizeOf(VM),
                     @Buf,
                     SizeOf(Buf) - 10,
                     NumRead,
                     nil);

     // If we read something, write it to the user
     if (NumRead > 0) then
     begin
          ComObj^.Com_SendBlock(Buf, NumRead, NumWritten);
     end;
end;

// Check to see if the program has anything to send to the user
procedure CheckRxNT;
var
  Buf: TBuf;
  BufPtr: Integer;
  I: Integer;
  MsgCount: Cardinal;
  NextSize: Cardinal;
  NumRead: Cardinal;
  NumWritten: LongInt;
begin
     BufPtr := 0;

     // Check to see how many messages are waiting
     if GetMailSlotInfo(ReadSlot,
                        nil,
                        NextSize,
                        @MsgCount,
                        nil) then
     begin
          // If a message is waiting, get it
          if MsgCount > 0 then
          begin
               for I := 1 to MsgCount do
               begin
                    // Read the next message
                    if ReadFile(ReadSlot,
                                Buf[BufPtr],
                                SizeOf(Buf) - BufPtr - 10,
                                NumRead,
                                nil) then
                    begin
                         Inc(BufPtr, NumRead);

                         // If we have filled the buffer, break
                         if (BufPtr >= SizeOf(Buf) - 10) then
                            Break;
                    end;
               end;
          end;
     end;

     // If we read something, write it to the user
     if (BufPtr > 0) then
        ComObj^.Com_SendBlock(Buf, BufPtr, NumWritten);
end;

// Check to see if the program has anything to send to the user
procedure CheckRx;
begin
     if (OS = WIN9X) then
        CheckRx9x
     else
     if (OS = WINNT) then
        CheckRxNT;
end;

(*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-*)

// Check to see if the user has anything to send to the program
procedure CheckTx9x;
var
  Buf: TBuf;
  InUsed: Integer;
  NumRead: Integer;
  OutBuf: Cardinal;
  OutBufRead: Cardinal;
begin
     if (ComObj^.Com_CharAvail) then
     begin
          // See what the comm buffers look like
          InUsed := ComObj^.InBuffer.BufUsed;

          if (InUsed > SizeOf(Buf) - 10) then
             InUsed := SizeOf(Buf) - 10;

          // Read a block of text
          ComObj^.Com_ReadBlock(Buf[4], InUsed, NumRead);
          Move(VM, Buf[0], SizeOf(VM));
          InUsed := InUsed  + SizeOf(VM);

          // Write the text to the program
          DeviceIOControl(VXD,
                          SBBSEXEC_IOCTL_WRITE,
                          @Buf,
                          InUsed,
                          @OutBuf,
                          SizeOf(OutBuf),
                          OutBufRead,
                          nil);
     end;
end;

// Check to see if the user has anything to send to the program
procedure CheckTxNT;
var
  Buf: TBuf;
  InUsed: LongInt;
  NumRead: LongInt;
  NumWritten: Cardinal;
  S: String;
begin
     // If the user has pressed a key
     if (ComObj^.Com_CharAvail) then
     begin
          // See what the comm buffers look like
          InUsed := ComObj^.InBuffer.BufUsed;

          if (InUsed > SizeOf(Buf) - 10) then
             InUsed := SizeOf(Buf) - 10;

          // Read a block of text
          ComObj^.Com_ReadBlock(Buf, InUsed, NumRead);

          // If our writeslot doesnt exist yet, create it
          if (WriteSlot = INVALID_HANDLE_VALUE) then
          begin
               // Create A Write Mail Slot
               S := '\\.\mailslot\sbbsexec\wr' + IntToStr(Node) + #0;
               WriteSlot := CreateFile(@S[1],
                                       GENERIC_WRITE,
                                       FILE_SHARE_READ,
                                       nil,
                                       OPEN_EXISTING,
                                       FILE_ATTRIBUTE_NORMAL,
                                       0);
          end;

          // Write the text to the program
          if Not(WriteSlot = INVALID_HANDLE_VALUE) then
             WriteFile(WriteSlot,
                       Buf,
                       NumRead,
                       NumWritten,
                       nil);
     end;
end;

// Check to see if the user has anything to send to the program
procedure CheckTx;
begin
     if (OS = WIN9X) then
        CheckTx9x
     else
     if (OS = WINNT) then
        CheckTxNT;
end;

(*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-*)

// Free the open handles
procedure CleanUp9x;
begin
     if (StartEvent <> INVALID_HANDLE_VALUE) then CloseHandle(StartEvent);
     if (VXD <> INVALID_HANDLE_VALUE) then CloseHandle(VXD);
end;

// Free the open handles
procedure CleanUpNT;
begin
     if (HungUpEvent <> INVALID_HANDLE_VALUE) then CloseHandle(HungUpEvent);
     if (ReadSlot <> INVALID_HANDLE_VALUE) then CloseHandle(ReadSlot);
     if (WriteSlot <> INVALID_HANDLE_VALUE) then CloseHandle(WriteSlot);
end;

// Free the open handles
procedure CleanUp;
begin
     if (OS = WIN9X) then
        CleanUp9x
     else
     if (OS = WINNT) then
        CleanUpNT;

     // Delete .ENV and .RET files
     DeleteFile(PChar(EnvFile));
     DeleteFile(PChar(RetFile));
end;

(*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-*)

// Initialize our program
function Init9x: Boolean;
var
  Kernel32: THandle;
  NumRead: Cardinal;
  OpenVxDHandle: TOpenVxDHandle;
  S: String;
begin
     // Assume initialization will fail
     Result := False;

     S := '\\.\' + ExtractFilePath(ParamStr(0)) + 'SBBSEXEC.VXD' + #0;
     VXD := CreateFile(@S[1],
                       0,
                       0,
                       nil,
                       CREATE_NEW,
                       FILE_FLAG_DELETE_ON_CLOSE,
                       0);
     if (VXD = INVALID_HANDLE_VALUE) then
     begin
          WriteLn('Failed To Create VXD');
          Exit;
     end;

     StartEvent := CreateEvent(nil,
                               True,
                               False,
                               nil);
     if (StartEvent = 0) then
     begin
          WriteLn('Failed To Create StartEvent');
          Exit;
     end;

     Kernel32 := LoadLibrary('KERNEL32.DLL');
     if (Kernel32 = 0) then
     begin
          WriteLn('Failed To Load KERNEL32.DLL');
          Exit;
     end else
     begin
          OpenVxDHandle := GetProcAddress(Kernel32, 'OpenVxDHandle');
          FreeLibrary(Kernel32);
     end;

     Start.Event := OpenVxDHandle(StartEvent);
     Start.Mode := SBBSEXEC_MODE_FOSSIL;

     if Not(DeviceIOControl(VXD,
                            SBBSEXEC_IOCTL_START,
                            @Start,
                            SizeOf(Start),
                            nil,
                            0,
                            NumRead,
                            nil)) then
     begin
          WriteLn('Failed To SBBSEXEC_IOCTL_START');
          Exit;
     end;

     // If we get here, initialization was successful
     Result := True;
end;

// Initialize our program
function InitNT: Boolean;
var
  S: String;
begin
     // Assume initialization will fail
     Result := False;

     // Initialize variables
     HungUpEvent := INVALID_HANDLE_VALUE;
     ReadSlot := INVALID_HANDLE_VALUE;
     WriteSlot := INVALID_HANDLE_VALUE;

     // Create a hungup event for when user drops carrier
     S := 'sbbsexec_hungup' + IntToStr(Node) + #0;
     HungUpEvent := CreateEvent(nil,
                                True,
                                False,
                                @S[1]);
     if (HungUpEvent = INVALID_HANDLE_VALUE) then
     begin
          WriteLn('Failed To Create HungUp Event');
          Exit;
     end;

     // Create a read mail slot
     S := '\\.\mailslot\sbbsexec\rd' + IntToStr(Node) + #0;
     ReadSlot := CreateMailSlot(@S[1],
                                XTRN_IO_BUF_LEN,
                                0,
                                nil);
     if (ReadSlot = INVALID_HANDLE_VALUE) then
     begin
          WriteLn('Failed To Create Read Mail Slot');
          Exit;
     end;

     // If we get here, initialization was successful
     Result := True;
end;

// Initialize our program
function Init: Boolean;
var
  F: Text;
  OSVer: TOSVersionInfo;
begin
     // Get OS version
     OSVer.dwOSVersionInfoSize := SizeOf(TOSVersionInfo);
     if (GetVersionEx(OSVer)) then
        OS := OSVer.dwPlatformID;

     if (OS = WIN9X) then
        Result := Init9x
     else
     if (OS = WINNT) then
        Result := InitNT
     else
        Result := False;

     // Initialize filename variables
     EnvFile := ExtractShortPathName(ExtractFilePath(ParamStr(0))) + 'DOSXTRN' + IntToStr(Node) + '.ENV';
     RetFile := ExtractShortPathName(ExtractFilePath(ParamStr(0))) + 'DOSXTRN' + IntToStr(Node) + '.RET';

     // Create temporary environment file
     Assign(F, EnvFile);
     {$I-}Rewrite(F);{$I+}
     if (IOResult = 0) then
     begin
          WriteLn(F, Command);
          Close(F);
     end;
end;

(*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-*)

// Parse the command line
procedure ParseCommandLine;
var
  I: Integer;
begin
     Command := '';
     LBY := 50;
     Node := -1;
     Sock := -1;

     if (ParamCount >= 5) then
     begin
          Node := StrToIntDef(ParamStr(1), -1);
          Sock := StrToIntDef(ParamStr(2), -1);
          LBY := StrToIntDef(ParamStr(3), 50);
          Command := ParamStr(4);
          for I := 5 to ParamCount do
              Command := Command + ' ' + ParamStr(I);
     end;

     // If missing important parameters, show syntax
     if (ParamCount < 4) or (Node = -1) or (Sock = -1) then
     begin
          WriteLn('USAGE:');
          WriteLn('  MSYNCFOS.EXE <node> <socket> <loops before yield> <command>');
          WriteLn;
          WriteLn('  <node>                 The node number');
          WriteLn('  <socket>               The socket handle');
          WriteLn('  <loops before yield>   The number of loops before yielding (default: 50)');
          WriteLn('  <command>              The command you want mSyncFos to execute');
          WriteLn;
          WriteLn('EXAMPLE:');
          WriteLn('  MSYNCFOS.EXE 1 1980 50 C:\LORD\START.BAT 1');
          WriteLn;
          WriteLn('NOTE:');
          WriteLn('  A low <loops before yield> value will decrease CPU usage, but increase in');
          WriteLn('  game delays.  Conversely, a high <loops before yield> value will increase');
          WriteLn('  CPU usage, but make games seem much more responsive');
          Halt;
     end;
end;

(*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-*)

// Run the dos fossil program
procedure RunProgram;
var
  CreationFlags: Cardinal;
  FullCmdLine: String;
  NumRead: Cardinal;
  PI: TProcessInformation;
  SI: TStartupInfo;
  RetVal: Cardinal;
begin
     // Initialize
     if Not(Init) then
     begin
          CleanUp;
          Exit;
     end;

     FillChar(SI, SizeOf(TStartupInfo), 0);
     SI.CB := SizeOf(TStartupInfo);

     // Start external program
     if (OS = WIN9X) then
     begin
          FullCmdLine := ExtractFilePath(ParamStr(0)) + 'DOSXTRN.EXE ' + EnvFile + ' 95 ' + IntToStr(Node) + ' ' + IntToStr(SBBSEXEC_MODE_FOSSIL) + ' ' + IntToStr(LBY) + #0;
          CreationFlags := CREATE_NEW_CONSOLE;
          SI.dwFlags := STARTF_USESHOWWINDOW;
          SI.wShowWindow := SW_MINIMIZE;
     end else
     begin
          FullCmdLine := ExtractFilePath(ParamStr(0)) + 'DOSXTRN.EXE ' + EnvFile + ' NT ' + IntToStr(Node) + ' ' + IntToStr(SBBSEXEC_MODE_FOSSIL) + ' ' + IntToStr(LBY) + #0;
          CreationFlags := CREATE_SEPARATE_WOW_VDM;
     end;

     if Not(CreateProcess(nil,
                          @FullCmdLine[1],
                          nil,
                          nil,
                          True,
                          CreationFlags,
                          nil,
                          nil,
                          SI,
                          PI)) then
     begin
          WriteLn('Failed To Create Process');
          CleanUp;
          Exit;
     end;

     // If Win9x, wait for new VM to start
     if (OS = WIN9X) then
     begin
          RetVal := WaitForSingleObject(StartEvent, 5000);
          if (RetVal <> WAIT_OBJECT_0) then
          begin
               WriteLn('Failed To Load New VM');
               TerminateProcess(PI.hProcess, 0);
          end;

          CloseHandle(StartEvent);

          DeviceIOControl(VXD,
                          SBBSEXEC_IOCTL_COMPLETE,
                          nil,
                          0,
                          @VM,
                          SizeOf(VM),
                          NumRead,
                          nil);
     end;

     // Loop until something happens
     while (True) do
     begin
          // Check for dropped carrier
          if Not(CheckCarrier) then
          begin
               TerminateProcess(PI.hProcess, 0);
               Sleep(500);
          end;

          // Check if program has terminated
          if (GetExitCodeProcess(PI.hProcess, RetVal)) then
          begin
               // If program is not active, break
               if (RetVal <> STILL_ACTIVE) then
                  Break;

               // Check to see if data is available for reading/writing
               CheckRx;
               CheckTx;
          end;

          Sleep(10);
     end;

     // Stop VxD
     if (OS = WIN9X) then
     begin
          DeviceIOControl(VXD,
                          SBBSEXEC_IOCTL_STOP,
                          @VM,
                          SizeOf(VM),
                          nil,
                          0,
                          NumRead,
                          nil);
     end;

     // Clean up
     if (RetVal = STILL_ACTIVE) then
        TerminateProcess(PI.hProcess, 0);

     CloseHandle(PI.hProcess);
     CloseHandle(PI.hThread);
     CleanUp;
end;

(*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-+-*-*)

begin
  // Show copyright notice
  WriteLn;
  WriteLn(PROG_VER);
  WriteLn('FOSSIL code written by Rob Swindell (www.synchro.net).');
  WriteLn('SyncFos is (c)2001-2002 by Maarten Bekers (www.elebbs.com)');
  WriteLn('mSyncFos is (c)2003 by Rick Parrish (www.mannsoft.ca)');
  WriteLn;

  // Parse command line
  ParseCommandLine;

  // Set up com object
  ComObj := New(PTelnetObj, Init);
  ComObj^.Com_SetDataProc(@ComReadProc, @ComWriteProc);
  ComObj^.DontClose := True;
  ComObj^.Com_OpenQuick(Sock);

  // Run the program
  RunProgram;

  // Destroy com object
  Dispose(ComObj, Done);
end.
